import java.util.ArrayList;
import java.util.Arrays;

public class Algorithms {

    int pheromonesStrength = -1;
    String antStartStrategy = "";
    int pheromonesDecayStrength = -1;
    String algorithmType;
    boolean resetPheromones = false;
    int maxRuntimeInSeconds = -1;
    int numberOfAnts = -1;
    int numberOfTaboos = -1;
    public Algorithms(){}
    public Algorithms(String algorithmType, String[] optionalParams) throws Exception{
        this.algorithmType = algorithmType;
        try {
            this.maxRuntimeInSeconds = Integer.valueOf(optionalParams[0]);
        } catch (Exception e) {
            
        }
        //checks if algorithm specific parameters are feasible and assigns them
        switch (algorithmType) {
            case "none":
                this.algorithmType = "basic_local_search";
                break;
            case "ant_colony":
                try {
                    this.numberOfAnts = Integer.valueOf(optionalParams[1]);
                    this.pheromonesStrength = Integer.valueOf(optionalParams[2]);
                    this.pheromonesDecayStrength = Integer.valueOf(optionalParams[3]);
                    if(Integer.valueOf(optionalParams[4]) == 0){
                        this.resetPheromones = false;
                    } else {
                        this.resetPheromones = true;
                    }
                    this.antStartStrategy = optionalParams[4];
                } catch (Exception e) {
                    
                }
                if(maxRuntimeInSeconds <= 0)
                    throw new Exception();
                if(this.pheromonesStrength <= 0)
                    System.out.println("Using standard value for pheromone strength.");
                if(this.antStartStrategy.equals(""))
                    System.out.println("Using standard value for ant start strategy.");
                if(this.pheromonesDecayStrength <= 0)
                    System.out.println("Using standard value for pheromone decay strength.");
                if(this.numberOfAnts <= 0)
                    System.out.println("Using standard value for number of ants.");
                break;
            case "taboo_search":
                try {
                    this.numberOfTaboos = Integer.valueOf(optionalParams[1]);
                } catch (Exception e) {
                    
                }
                if(maxRuntimeInSeconds <= 0)
                    throw new Exception();
                if(this.numberOfTaboos <= 0)
                    System.out.println("Using standard value for number of Taboos."
                );
                break;
            default:
                break;
        }
    }

    public Solution generateSolution(Instance instance){
            switch (this.algorithmType) {
                case "greedy":
                    return generateInitialSolutionGreedy(instance);
                case "ant_colony":
                    //setting default parameters if needed
                    if(this.numberOfAnts <= 0)
                        numberOfAnts = Math.multiplyExact(instance.getDimension(),5);
                    if(this.pheromonesDecayStrength <= 0)
                        pheromonesDecayStrength = 1;
                    if(this.pheromonesStrength <= 0)
                        pheromonesStrength = 10;
                    if(this.antStartStrategy.equals(""))
                        antStartStrategy = "even";
                    return generateInitialSolutionAnt(instance);
                default:
                    return null;
            }
    }

    public Solution generateSolution(Solution solution){
        switch (this.algorithmType) {
            case "basic_local_search":
                return localSearch(solution);
            case "greedy":
                return generateInitialSolutionGreedy(solution.instance);
            case "taboo_search":
                //setting default parameters if needed
                if(this.numberOfTaboos <= 0)
                    numberOfTaboos = solution.instance.getDimension()/3;
                return tabooSearch(solution);
            case "ant_colony":
                return generateSolution(solution.instance);
            default:
                return localSearch(solution);
        }
    }

    private Solution tabooSearch(Solution solution) {
        int[] state = {2,0,0,0};
        int newBestCount = 0;
        int time =0;
        Solution currentSolution = solution;
        Solution bestSolutionSeen = solution;
        
        long start = System.currentTimeMillis();
        long end = start + this.maxRuntimeInSeconds * 1000;
        Solution possibleBetterNeighbor = getOtherNeighbor(state, bestSolutionSeen);
        Solution currentBestNeighbor = possibleBetterNeighbor;

        int currentBestNeighborNodeChanged = 2;
        
        ArrayList<Integer> taboos = new ArrayList<>();
    
        
        while(maxRuntimeInSeconds > 0 && System.currentTimeMillis()<end){
            if(System.currentTimeMillis()>start+1000 && time == 0){
                try{System.out.println("1 Sekunde: " + bestSolutionSeen.getCost());}catch(Exception e){}
                time++;
            }
            if(System.currentTimeMillis()>start+10000 && time == 1){
                try{System.out.println("10 Sekunde: " + bestSolutionSeen.getCost());}catch(Exception e){}
                time++;
            }
            if(System.currentTimeMillis()>start+60000 && time == 2){
                try{System.out.println("60 Sekunde: " + bestSolutionSeen.getCost());}catch(Exception e){}
                time++;
            }
            try {
                if(possibleBetterNeighbor.getCost() < currentBestNeighbor.getCost()){
                    currentBestNeighbor = possibleBetterNeighbor;
                    currentBestNeighborNodeChanged = state[0];
                }
            } catch (Exception e) {
                
            
            }
            
            //permutate state to get another neighbor which is not taboo
            state[0]++;
            while(taboos.contains(state[0])){
                state[0]++;
            }
            if(state[0] > bestSolutionSeen.instance.getDimension()){
                //System.out.println("reset extracted node");
                state[0] = 2;
                state[2]++;
                if(state[2] > bestSolutionSeen.tours[state[1]].length-1){
                    //System.out.println("reset position");
                    state[2] = 0;
                    state[1]++;
                    if(state[1] > bestSolutionSeen.tours.length-1){
                        //System.out.println("reset tours");
                        state[1] = 0;
                        state[3]++;
                        //choose minimal after all neighbors are seen, update taboos, best solution see so far and current solution
                        if(state[3] == 2){
                            currentSolution = currentBestNeighbor;
                            //System.out.println("Made a step, taboos: "+taboos.size());
                            try {
                                if(currentSolution.getCost() < bestSolutionSeen.getCost()){
                                    bestSolutionSeen = currentSolution;
                                    
                                    System.out.println("Found new best solution "+(++newBestCount) + ": All "+taboos.size()+" taboos dropped");
                                    taboos.clear();
                                } else {
                                    taboos.add(currentBestNeighborNodeChanged);
                                    if(taboos.size() > numberOfTaboos){
                                        taboos.remove(0);
                                    }
                                }
                            } catch (Exception e) {
                               
                            }
                            
                            int[] help = {2,0,0,0};
                            state = help;
                            currentBestNeighbor = getOtherNeighbor(state, currentSolution);
                        }
                    }
                }
            }
            possibleBetterNeighbor = getOtherNeighbor(state, currentSolution);
        }
        System.out.println("terminated by time limit");
        return bestSolutionSeen;
    }

    private Solution generateInitialSolutionAnt(Instance instance) {
        Solution bestSeenSolution = null;
        int time = 0;
        int newBestSolutionCount = 0;
        int[][] pheromonesOnEdge = new int[instance.getDimension()+1][instance.getDimension()+1];
        //initialize all edges with phermone score 1
        for (int i = 0; i < pheromonesOnEdge.length; i++) {
            for (int j = 0; j < pheromonesOnEdge[i].length; j++) {
                pheromonesOnEdge[i][j] = 1;
            }
        }
        
        long start = System.currentTimeMillis();
        long end = start + this.maxRuntimeInSeconds * 1000;

        while(maxRuntimeInSeconds > 0 && System.currentTimeMillis()<end){
            //if optional parameter is set, reset pheromones inbetween major rounds
            if(System.currentTimeMillis()>start+1000 && time == 0){
                try{System.out.println("1 Sekunde: " + bestSeenSolution.getCost());}catch(Exception e){}
                time++;
            }
            if(System.currentTimeMillis()>start+10000 && time == 1){
                try{System.out.println("10 Sekunde: " + bestSeenSolution.getCost());}catch(Exception e){}
                time++;
            }
            if(System.currentTimeMillis()>start+60000 && time == 2){
                try{System.out.println("60 Sekunde: " + bestSeenSolution.getCost());}catch(Exception e){}
                time++;
            }
            if(resetPheromones) {
                for (int i = 0; i < pheromonesOnEdge.length; i++) {
                    for (int j = 0; j < pheromonesOnEdge[i].length; j++) {
                        pheromonesOnEdge[i][j] = 1;
                    }
                }
            }
                
            ArrayList<Ant> ants = new ArrayList<Ant>();
            for (int i = 0; i < numberOfAnts; i++) {
                switch (antStartStrategy) {
                    case "depot":
                        ants.add(new Ant(instance));
                        break;
                    case "even":
                    default:
                        ants.add(new Ant(instance, i+1));
                        break;
                }
                
            }
            while(!antsAreAllDone(ants)){
                ArrayList<int[]> walkedEdges = new ArrayList<int[]>();
                //let ants do one minor round
                for (Ant ant : ants) {
                    int[] walkedEdge = ant.makeStepInMinorRound(pheromonesOnEdge);
                    walkedEdges.add(walkedEdge);
                }
                //update pheromones after minor round
                for (int[] walkedEdge : walkedEdges) {
                    if(walkedEdge != null){
                        if(walkedEdge[0] < walkedEdge[1]){
                            pheromonesOnEdge[walkedEdge[1]][walkedEdge[0]] += pheromonesStrength;
                        } else {
                            pheromonesOnEdge[walkedEdge[0]][walkedEdge[1]] += pheromonesStrength;
                        }
                    }
                }
                //let pheromones decay (not under 1)
                for (int i = 0; i < pheromonesOnEdge.length; i++) {
                    for (int j = 0; j < pheromonesOnEdge[i].length; j++) {
                        pheromonesOnEdge[i][j] -= pheromonesDecayStrength;
                        if(pheromonesOnEdge[i][j]<1){
                            pheromonesOnEdge[i][j] = 1;
                        }
                    }
                }
            }
            //check if ant has discovered new best solution
            for (Ant ant : ants) {
                Solution antsSolution = ant.toSolution();
                
                if(bestSeenSolution == null){
                    System.out.println("Found new best solution "+(++newBestSolutionCount));
                    bestSeenSolution = antsSolution;
                } else {
                    try {
                        if(bestSeenSolution.getCost() > antsSolution.getCost()){
                            System.out.println("Found new best solution "+(++newBestSolutionCount));
                            bestSeenSolution = antsSolution;
                        }
                    } catch (Exception e) {
                       
                    }
                }
            
                
            }
        }

        return bestSeenSolution;
    }

    private boolean antsAreAllDone(ArrayList<Ant> ants){
        boolean allDone = true;
        for (Ant ant : ants) {
            if(!ant.done){
                allDone = false;
                break;
            }
        }
        return allDone;
    }

    //primitive greedy Approach
    private Solution generateInitialSolutionGreedy(Instance instance){
        Solution solution = new Solution(instance);
        ArrayList<ArrayList<Node>> tours = new ArrayList<ArrayList<Node>>();
        ArrayList<Integer> neededNodes = new ArrayList<Integer>();
        for (int index = 2; index < instance.getDimension()+1; index++) {
            neededNodes.add(index);
        }
        ArrayList<Node> currentTour = new ArrayList<Node>();
        int lastNodeNumber = instance.getDepot().number;
        int currentBestCost = -1;
        int currentBestNodeNumber = 1;
        int currentDemand = 0;
        while(neededNodes.size()>0){
            for (Integer i : neededNodes) {
                if(i < lastNodeNumber){
                    if(currentBestCost == -1 || instance.getEdgeWeights()[lastNodeNumber][i] < currentBestCost){
                        currentBestCost = instance.getEdgeWeights()[lastNodeNumber][i];
                        currentBestNodeNumber = i;
                    }
                }else{
                    if(currentBestCost == -1 || instance.getEdgeWeights()[i][lastNodeNumber] < currentBestCost){
                        currentBestCost = instance.getEdgeWeights()[i][lastNodeNumber];
                        currentBestNodeNumber = i;
                    }
                }
            }
            if(currentDemand + instance.getNodes()[currentBestNodeNumber].demand < instance.getTruckCapacity()){
                currentDemand += instance.getNodes()[currentBestNodeNumber].demand;
                if(instance.getNodes()[currentBestNodeNumber] == null){
                    System.out.println("k");
                }else{
                    currentTour.add(instance.getNodes()[currentBestNodeNumber]);
                }
                
                int help = currentBestNodeNumber;
                neededNodes.removeIf(s -> s.equals(help));
                lastNodeNumber = currentBestNodeNumber;
                currentBestCost = -1;
                currentBestNodeNumber = 1;
            }else{
                tours.add(new ArrayList<Node>(currentTour));
                currentTour = new ArrayList<Node>();
                currentDemand = 0;

                lastNodeNumber = instance.getDepot().number;
                currentBestCost = -1;
                currentBestNodeNumber = 1;
                
            }
        }
        tours.add(new ArrayList<Node>(currentTour));
        Node[][] convertedTours = new Node[tours.size()][];
        int i =0;
        for (ArrayList<Node> tour : tours) {
            convertedTours[i] = new Node[tour.size()];
            convertedTours[i] = tour.toArray(convertedTours[i]);
            i++;
        }
        solution.tours = convertedTours;
        return solution;
    }

    //local Search 
    private Solution localSearch(Solution solution){
        Solution currentBestSolution = solution;
        int[] state = {2,0,0,0};
        Solution possibleBetterSolution = getOtherNeighbor(state, currentBestSolution);
        int newBestSolutionCount = 0;
        long start = System.currentTimeMillis();
        long end = start + this.maxRuntimeInSeconds * 1000;
        //tries all neighbors as possibleBetterSolution, if better solution found, change currentBestSolution and search from there 
        //terminates when it cannot find a better solution in its neighbarhood
        while (possibleBetterSolution != null) {
            
            //stop if max execution time has exceeded
            if(maxRuntimeInSeconds > 0 && System.currentTimeMillis()>end){
                System.out.println("terminated by time limit");
                break;
            }
            try {
                if(possibleBetterSolution.getCost() < currentBestSolution.getCost()){
                    currentBestSolution = possibleBetterSolution;
                    int[] help = {2,0,0,0};
                    state = help;
                    possibleBetterSolution = getOtherNeighbor(state, currentBestSolution);
                    System.out.println("Found new best solution "+(++newBestSolutionCount));
                }else{
                    //permutate state to get another neighbor
                    state[0]++;
                    if(state[0] > currentBestSolution.instance.getDimension()){
                        //System.out.println("reset extracted node");
                        state[0] = 2;
                        state[2]++;
                        if(state[2] > currentBestSolution.tours[state[1]].length-1){
                            //System.out.println("reset position");
                            state[2] = 0;
                            state[1]++;
                            if(state[1] > currentBestSolution.tours.length-1){
                                //System.out.println("reset tours");
                                state[1] = 0;
                                state[3]++;
                                //stop if all neighbors are worse/local optimum is reached
                                if(state[3] == 2){
                                    System.out.println("terminated by local optimum");
                                    break;
                                }
                            }
                        }
                    } 
                    possibleBetterSolution = getOtherNeighbor(state, currentBestSolution);
                }
            } catch (Exception e) {
                
            }
            
            
        }
       
        
        return currentBestSolution;
    }

    //State[0] = Number Of Extract Node, State[1] = Number Of Insert Tour, State[2] = Number Of Insert Position, State[3] = 0/1 tryNewTour,
    //finds a specific neighbor in the neighborhood relationship
    private Solution getOtherNeighbor(int[] state, Solution solution){
        Solution possibleBetterSolution = solution.copy();
        //removing a node
        for (Node[] tour : possibleBetterSolution.tours) {
            ArrayList<Node> list = new ArrayList<Node>(Arrays.asList(tour));
            list.removeIf(s -> s != null && s.number == state[0]);
            tour = list.toArray(tour);
        }
        //inserting it elsewhere
        Node searchedNode = new Node(-1, -1, -1);
        searchedNode.number = -1;
            for (Node node : solution.instance.getNodes()) {
                if(node != null){
                    if(node.number == state[0]) {searchedNode = node; break;}
                }
            }
        if(state[3]==1){
            ArrayList<Node[]> list = new ArrayList<Node[]>(Arrays.asList(possibleBetterSolution.tours));
            Node[] newTour = {searchedNode};
            list.add(newTour);
            possibleBetterSolution.tours = list.toArray(possibleBetterSolution.tours);
        }else {
            ArrayList<Node> list = new ArrayList<Node>(Arrays.asList(possibleBetterSolution.tours[state[1]]));
            if(searchedNode == null){
                System.out.println("a");
            }else{
                try {
                    list.add(state[2], searchedNode);
                } catch (Exception e) {
                    // TODO: handle exception
                    System.out.println("");
                }
                
            }
            
            possibleBetterSolution.tours[state[1]] = list.toArray(possibleBetterSolution.tours[state[1]]);
        }
        

        return possibleBetterSolution;
    }
}
