 import java.util.ArrayList;
import java.util.Arrays;

public class Ant {
    Instance instance;
    Node currentNode;
    ArrayList<Node> currentTour = new ArrayList<Node>();
    ArrayList<ArrayList<Node>> tours = new ArrayList<ArrayList<Node>>();
    ArrayList<Node> notVisitedNodes;
    boolean done = false;

    public Ant(Instance instance){
        this.instance = instance;
        notVisitedNodes = new ArrayList<Node>(Arrays.asList(instance.getNodes()));
        notVisitedNodes.remove(instance.getDepot());
        currentNode = instance.getDepot();
        notVisitedNodes.remove(0);
    }

    public Ant(Instance instance, int startNodeNumber){
        Node currentNode;
        currentNode = instance.getDepot();
        this.instance = instance;
        notVisitedNodes = new ArrayList<Node>(Arrays.asList(instance.getNodes()));
        for (Node node : notVisitedNodes) {
            if(node != null && node.number == (startNodeNumber%instance.getDimension())+1){
                currentNode = node;
                break;
            }
        }
        notVisitedNodes.remove(instance.getDepot());
        notVisitedNodes.remove(0);
        notVisitedNodes.remove(currentNode);
        this.currentNode = currentNode;
        currentTour.add(currentNode);
        
    }

    public int[] makeStepInMinorRound(int[][] pheromonesOnEdge){
        //if ant is done with tour return null, else return edge, that she walked
        if(this.done){
            return null;
        }
        Node nextNode = null;
            
        double currentProbability = 0;
        ArrayList<double[]> probabilities = new ArrayList<double[]>();
        if(currentNode.depot){
            currentTour = new ArrayList<Node>();
        } else {
            double distance;
            double pheromones;
            if(currentNode.number < instance.getDepot().number){
                distance = instance.getEdgeWeights()[instance.getDepot().number][currentNode.number];
                pheromones = pheromonesOnEdge[instance.getDepot().number][currentNode.number];
            } else {
                distance = instance.getEdgeWeights()[currentNode.number][instance.getDepot().number];
                pheromones = pheromonesOnEdge[instance.getDepot().number][currentNode.number];
            }
            double probability = pheromones / distance;
            double[] probabilityInterval= new double[3];
            probabilityInterval[0] = currentProbability;
            probabilityInterval[1] = currentProbability+probability;
            probabilityInterval[2] = instance.getDepot().number;
            probabilities.add(probabilityInterval);
            currentProbability += probability;
        }
        for (Node node : notVisitedNodes) {
            if(node.demand + sumOfDemands(currentTour) <= instance.getTruckCapacity()){
                double distance;
                double pheromones;
                if(currentNode.number < node.number){
                    distance = instance.getEdgeWeights()[node.number][currentNode.number];
                    pheromones = pheromonesOnEdge[node.number][currentNode.number];
                } else {
                    distance = instance.getEdgeWeights()[currentNode.number][node.number];
                    pheromones = pheromonesOnEdge[node.number][currentNode.number];
                }
                double probability = pheromones / (distance+1);
                double[] probabilityInterval= new double[3];
                probabilityInterval[0] = currentProbability;
                probabilityInterval[1] = currentProbability+probability;
                probabilityInterval[2] = node.number;
                probabilities.add(probabilityInterval);
                currentProbability += probability;
            }
        }
        double randomNumber = Math.random()*currentProbability;
        int nextNodeNumber = -1;
        for (double[] interval : probabilities) {
            if(randomNumber >= interval[0] && randomNumber < interval[1]){
                 nextNodeNumber = (int)interval[2];
            }
        }

        if(nextNodeNumber == instance.getDepot().number){
            tours.add(currentTour);
            nextNode = instance.getDepot();
            if(notVisitedNodes.size() == 0){
                this.done = true;
            }
        } else {
            for (Node node : notVisitedNodes) {
                if(node.number == nextNodeNumber){
                    currentTour.add(node);
                    nextNode = node;
                    notVisitedNodes.remove(node);
                    break;
                }
            }
        }

        
        int[] walkedEdge = {currentNode.number,nextNode.number};
        currentNode = nextNode;
        
        return walkedEdge;
    }

    private int sumOfDemands(ArrayList<Node> currentTour){
        int demand = 0;
        for (Node node : currentTour) {
            demand += node.demand;
        }
        return demand;
    }

    public Solution toSolution(){
        Solution sol = new Solution(instance);
        Node[][] tours = new Node[this.tours.size()][];
        int i = 0;
        for (ArrayList<Node> tour : this.tours) {
            tours[i] = new Node[tour.size()];
            tours[i] = tour.toArray(tours[i]);
            i++;
        }
        sol.tours = tours;
        return sol;
    }
}
