import java.io.File;
import java.io.FileNotFoundException;  // Import this class to handle errors
import java.util.Scanner; // Import the Scanner class to read text files

//represents one instance
public class Instance {
    
    private int truckCapacity;
    private int dimension;
    private String name;
    private int[][] edgeWeights;
    private Node[] nodes;
    private Node depot;
    private int optimalCost;

    public  Instance(){}
    public  Instance(String fileLocation) throws Exception{
        //import instance of cvrp from file at fileLocation and optimal solutioncost from sol file in same folder, lower Marix of distances
        try {
            File file = new File(fileLocation);
            Scanner scanner = new Scanner(file);
            while (scanner.hasNextLine()) {
                String data = scanner.nextLine();
                if(data.contains("NAME")) this.name = data.replace("NAME : ", "");
                else if(data.contains("DIMENSION")) this.dimension = Integer.valueOf(data.replace("DIMENSION : ", "")); 
                else if(data.contains("CAPACITY")) this.truckCapacity = Integer.valueOf(data.replace("CAPACITY : ", ""));  
                else if(data.contains("EDGE_WEIGHT_SECTION")){
                    data = scanner.nextLine();
                    edgeWeights = new int[this.dimension+1][this.dimension+1];
                    while(!data.equals("NODE_COORD_SECTION")){
                        String[] edgeweighStrings = data.split(" ");
                        for (int i = 0; i < edgeweighStrings.length;i++) {
                            edgeWeights[edgeweighStrings.length+1][i+1] = Integer.valueOf(edgeweighStrings[i]);
                        }
                        data = scanner.nextLine();
                    }
                    data = scanner.nextLine();
                    nodes = new Node[this.dimension+1];
                    while(!data.equals("DEMAND_SECTION")){
                        String[] nodeCoords = data.split(" ");
                        Integer number = Integer.valueOf(nodeCoords[0]);
                        nodes[number] = new Node(number, Integer.valueOf(nodeCoords[1]), Integer.valueOf(nodeCoords[2]));
                        data = scanner.nextLine();
                    }
                    data = scanner.nextLine();
                    while(!data.trim().equals("DEPOT_SECTION")){

                        try {
                            String[] nodeCoords = data.split(" ");
                            Integer number = Integer.valueOf(nodeCoords[0]);
                            nodes[number].setDemand(Integer.valueOf(nodeCoords[1]));
                            data = scanner.nextLine();
                        } catch (Exception e) {
                            break;
                        }
                        
                    }
                    data = scanner.nextLine();
                    while(!data.equals("-1")){
                        
                        
                        Integer number = Integer.valueOf(data.trim());
                        nodes[number].setDepot();
                        this.depot = nodes[number];
                        data = scanner.nextLine();
                    }
                }   
              }
              scanner.close();
        } catch (FileNotFoundException e) {
            System.err.println("An error occurred at import, please check if your .vrp file fulfills the readme-conditions and the path is correct.");
            throw new Exception();
        }
        try {
            File file = new File(fileLocation.replace(".vrp", ".sol"));
            Scanner scanner = new Scanner(file);
            while (scanner.hasNextLine()) {
                String data = scanner.nextLine();
                if(data.contains("Cost")) this.optimalCost = Integer.valueOf(data.replace("Cost", "").trim());
              }
              scanner.close();
        } catch (FileNotFoundException e) {
            System.out.println("No .sol file found for optimal solution.");
        }
    }

    public String getName(){
        return this.name;
    }
    public int getOptimalCost(){
        return this.optimalCost;
    }
    public int getDimension(){
        return this.dimension;
    }
    public int getTruckCapacity(){
        return this.truckCapacity;
    }
    public int[][] getEdgeWeights(){
        return this.edgeWeights;
    }
    public Node[] getNodes(){
        return this.nodes;
    }
    public Node getDepot(){
        return this.depot;
    }
}