import java.util.Arrays;
import java.util.ArrayList;
import java.util.HashSet;

//represents a feasible solution to a instance
public class Solution {
    Instance instance;
    Node[][] tours;
    private int cost;
    private boolean feasible = false;

    public Solution(Instance instance){
        this.instance = instance;
    }

    public boolean isFeasible() throws Exception{
        ArrayList<Node> allVisitedNodes = new ArrayList<Node>();
        for (Node[] tour : this.tours) {
            int tourCost = 0;
            int tourDemand = 0;
            Node lastNode = instance.getDepot();
            for (Node node : tour) {
                if(node != null){
                    tourDemand += node.demand;
                    allVisitedNodes.add(node);
                    if(lastNode.number > node.number){
                        tourCost += instance.getEdgeWeights()[lastNode.number][node.number];
                    } else {
                        tourCost += instance.getEdgeWeights()[node.number][lastNode.number];
                    }
                    lastNode = node;
                }
            }
            tourCost += instance.getEdgeWeights()[lastNode.number][instance.getDepot().number];
            //check if tour is in capacity limit
            if(tourDemand > instance.getTruckCapacity()) {
                throw new Exception("capacity problem ("+tourDemand + " greater " +instance.getTruckCapacity()+")");
            }
              
        }
        //check if nodes are duplicates
        if(allVisitedNodes.size() != new HashSet<Node>(allVisitedNodes).size()) {
            throw new Exception("duplicate problem");
        }
        //check if right number of nodes is in all tours
        if(allVisitedNodes.size() != instance.getDimension()-1) {
            throw new Exception("not all nodes problem");
        }

        return this.feasible = true;
    }

    public int getCost() throws Exception{
        isFeasible();
        this.cost = 0;
        for (Node[] tour : tours) {
            Node lastNode = instance.getDepot();
            for (Node node : tour) {
                try {
                    if(lastNode.number > node.number){
                        this.cost += instance.getEdgeWeights()[lastNode.number][node.number];
                    } else {
                        this.cost += instance.getEdgeWeights()[node.number][lastNode.number];
                    }
                    lastNode = node;
                } catch (Exception e) {
                   
                }
                
            }
            this.cost += instance.getEdgeWeights()[lastNode.number][instance.getDepot().number];
        }
        return this.cost;
    };
    public String toString(){
        String s = "name: "+instance.getName()+ "\n" + "cost: " + cost + "feasible: " + feasible + "\nTours: ";
        for (Node[] tour : tours) {
            s += "\n[";
            for (Node node : tour) {
                s+=node.number+",";
            }
            s += "]";
        }
        return s.replace(",]", "]");
    }

    //deepcopy of a solution
    public Solution copy(){
        Solution copy = new Solution(instance);
        copy.cost = this.cost;
        Node[][] toursCopy = new Node[tours.length][];
        for (int i = 0; i < tours.length; i++) {
            toursCopy[i] = Arrays.copyOf(tours[i], tours[i].length);
        }
        copy.tours = toursCopy;
        return copy;
    }

    public String toSol(){
        String rep = "";
        
        for (int i = 0; i < tours.length; i++) {
           rep += "Route #"+(i+1)+":"; 
           for (Node node : tours[i]) {
            try {
                rep += " "+node.number;
            } catch (Exception e) {
                
            }              
           } 
           rep += "\n";
        }
        try {
            rep += "Cost " + this.getCost();
        } catch (Exception e) {
            rep = "not feasible: " + e.getMessage() + "\n" + rep;
        }
        
        return rep;
    }
}